unit uMemMap;

{******************************************************************************
Initial Author   :    Scott Price
Initially Created:    06-Apri-2003
Language         :    Object Pascal
Compiler Version :    Delphi 3.02, 4.03, 5.01, 6.02 (UK)
Copyright        :     Information Technology Sector Limited, 1996-2002

Source Code Constraints:
========================
   Distribution, disclosure or modifications of this source code, outside of
   any licenced agreements with Information Technology Sector Limited, or
   removal of this notice constitutes a breach of any or all license agreements.

Operating Systems Tested On:
============================
   Microsoft Windows(R) 98, First Edition
   Microsoft Windows(R) 2000, Service Pack 2

Design Notes:
=============

VERSION  AUTHOR                DATE/TIME
===============================================================================
$history$
$nokeywords$
===============================================================================
NOTES:
 - Memory Mapping Sizes increase in 4k chunks.
===============================================================================
TODO:
******************************************************************************}


interface



uses
  { Borland Standard Units }
  Classes, Windows, SysUtils;


type
  { Exception Class Definitions }
  ERequirementsNotMet = class(Exception);


  { TMemoryMapping }
  { This class exists to wrap the elements of connecting to, or creating, memmory mapped shares }

  TMemoryMapping = class(TPersistent)
  private
    { Private Declarations }
    FMemMapIdent: String;
    FMemory: Pointer;
    FHandle: THandle;
    FSize: Integer;
    FAccess: DWORD;
    FCreated: Boolean;
    FLastError: DWORD;
    function GetConnected: Boolean;
    procedure SetMemMapID(const Value: String);
    procedure SetSize(const Value: Integer);
    function MappingSucceeded: Boolean;
  protected
    { Protected Declarations }
    { This is here to check for the contractual requirements }
    function CheckRequirements: Boolean;
    function GetMemory: Pointer;
  public
    { Public Declarations }
    constructor Create; virtual;
    destructor Destroy; override;
    function OpenMapping: Boolean; virtual;
    function OpenMappingOnly: Boolean; virtual;
    function CreateMapping: Boolean; virtual;
    function CloseMapping: Boolean; virtual;
    property Memory: Pointer read FMemory;
  published
    { Published Declarations }
    { These access rights are essential! }
    property Access: DWORD read FAccess write FAccess;
    property Created: Boolean read FCreated;
    property Connected: Boolean read GetConnected;
    property Handle: THandle read FHandle;
    property MemMapIdent: String read FMemMapIdent write SetMemMapID;
    property Size: Integer read FSize write SetSize;
    property LastError: DWORD read FLastError;
  end;




implementation




{ TMemoryMapping }

const
  scErrConnected = 'The object is already connected, and must have CloseMapping ' +
                   'called prior to changing property values!';
  scErrRequirements = 'The use of this class Requires a (Size > 0) AND ' +
                      '(Length(MemMapID) > 0).  Please check the property ' +
                      'values and try again!';
  scErrNotOpen = 'The Memory Mapping is not Connected!';


constructor TMemoryMapping.Create;
begin
  inherited Create;

  FMemMapIdent := '';
  FMemory := Nil;
  FHandle := 0;
  FSize := -1;
  { These access rights are essential!
    Incorrect settings of these access modes will deny you access to the memory }
  FAccess := FILE_MAP_ALL_ACCESS;
  FCreated := False;
  FLastError := 0;
end;

destructor TMemoryMapping.Destroy;
begin
  TRY
    if Connected then
      CloseMapping;
  FINALLY
    inherited Destroy;
  END;
end;

function TMemoryMapping.GetConnected: Boolean;
begin
  Result := MappingSucceeded;
end;

procedure TMemoryMapping.SetMemMapID(const Value: String);
begin
  if (FMemMapIdent <> Value) then
  begin
    if Connected then
      EInvalidOperation.Create(scErrConnected);

    FMemMapIdent := Value;
  end;
end;

procedure TMemoryMapping.SetSize(const Value: Integer);
begin
  if (FSize <> Value) then
  begin
    if Connected then
      EInvalidOperation.Create(scErrConnected);

    FSize := Value;
  end;
end;

function TMemoryMapping.MappingSucceeded: Boolean;
begin
  Result := ((FHandle <> 0) AND (Memory <> Nil));
end;

function TMemoryMapping.CheckRequirements: Boolean;
begin
  { We have to check the requirements before trying to attach to the memory
    mapping.  We have to make sure that all requirements are met prior to
    allowing further progress. }
  Result := ((Size > 0) AND (Length(FMemMapIdent) > 0));
end;

function TMemoryMapping.CreateMapping: Boolean;
begin
  { Try to create the Memory Mapping if we aren't already connected to one
    and have all the required settings }
  if Connected then
    raise EInvalidOperation.Create(scErrConnected);

  if (CheckRequirements = False) then
    raise ERequirementsNotMet.Create(scErrRequirements);

  FHandle := CreateFileMapping($FFFFFFFF, Nil, PAGE_READWRITE, 0, FSize,
               PChar(FMemMapIdent));

  if (FHandle = 0) then
    FLastError := GetLastError;

  FMemory := MapViewOfFile(FHandle, FAccess, 0, 0, 0);

  Result := MappingSucceeded;

  FCreated := Result;
end;

function TMemoryMapping.CloseMapping: Boolean;
begin
  { Try to close the Memory mapping if we are Connected }
  if (Connected = False) then
  begin
    Result := False;
    Exit;
  end;

  UnmapViewOfFile(FMemory);
  FMemory := Nil;

  CloseHandle(FHandle);
  FHandle := 0;

  Result := True;
end;

function TMemoryMapping.OpenMapping: Boolean;
begin
  { If not connected try to Open the memory mapping if it already exists.
    If no mapping already exists, create one }
  if Connected then
    raise EInvalidOperation.Create(scErrConnected);

  if (CheckRequirements = False) then
    raise ERequirementsNotMet.Create(scErrRequirements);

  FHandle := OpenFileMapping(FAccess, False, PChar(FMemMapIdent));

  if (FHandle = 0) then
    Result := CreateMapping
  else
  begin
    FMemory := MapViewOfFile(FHandle, FAccess, 0, 0, 0);

    Result := MappingSucceeded;
  end;
end;

function TMemoryMapping.OpenMappingOnly: Boolean;
begin
  { If we aren't already connected, then try to Open the Mapping.
    There is no need to try to create a mapping here }
  if Connected then
    raise EInvalidOperation.Create(scErrConnected);

  if (CheckRequirements = False) then
    raise ERequirementsNotMet.Create(scErrRequirements);

  FHandle := OpenFileMapping(FAccess, False, PChar(FMemMapIdent));

  { If the returned Zero, then the open failed! }
  if (FHandle <> 0) then
  begin
    FMemory := MapViewOfFile(FHandle, FAccess, 0, 0, 0);

    Result := MappingSucceeded;
  end
  else
    Result := False;
end;

function TMemoryMapping.GetMemory: Pointer;
begin
  Result := FMemory;
end;

end.
